__version__ = '0.0.1'

from ds4mllib.dataset import DataSet
from ds4mllib.attribute import Attribute
from ds4mllib.evaluator import BiFrame

# Default environment variables for data processing and analysis
# It's not recommend to modify these variables' values.
params = {
    'attribute.bins': 20
}
